"""
Flash search for cos / reverse cos
return matched peaks, spectral usage, reverse score
"""
import json
import multiprocessing
import pickle
from functools import reduce
from pathlib import Path
from typing import Union, List

import numpy as np

from _preprocess_ms2 import preprocess_ms2

np.seterr(divide='ignore', invalid='ignore')


class FlashCosCore:
    def __init__(self, path_data=None, max_ms2_tolerance_in_da=0.024, mz_index_step=0.0001,
                 peak_scale_k=8.0,
                 peak_intensity_power=1.0):
        """
        Initialize the search core. for CPU only.
        :param path_array: The path array of the index files.
        :param max_ms2_tolerance_in_da: The maximum MS2 tolerance used when searching the MS/MS spectra, in Dalton. Default is 0.024.
        :param mz_index_step:   The step size of the m/z index, in Dalton. Default is 0.0001.
                                The smaller the step size, the faster the search, but the larger the index size and longer the index building time.
        """
        self.mz_index_step = mz_index_step
        self._init_for_multiprocessing = False
        self.max_ms2_tolerance_in_da = max_ms2_tolerance_in_da
        self.peak_scale_k = peak_scale_k
        self.peak_intensity_power = peak_intensity_power

        self.total_spectra_num = 0
        self.total_peaks_num = 0
        self.index = []

        if path_data:
            self.path_data = Path(path_data)
        else:
            self.path_data = None

        self.index_names = [
            "all_ions_mz_idx_start",
            "all_ions_mz",
            "all_ions_intensity",
            "all_ions_spec_idx",
            "all_nl_mass_idx_start",
            "all_nl_mass",
            "all_nl_intensity",
            "all_nl_spec_idx",
            "all_ions_idx_for_nl",
        ]
        self.index_dtypes = {
            "all_ions_mz_idx_start": np.int64,
            "all_ions_mz": np.float32,
            "all_ions_intensity": np.float32,
            "all_ions_spec_idx": np.uint32,
            "all_nl_mass_idx_start": np.int64,
            "all_nl_mass": np.float32,
            "all_nl_intensity": np.float32,
            "all_nl_spec_idx": np.uint32,
            "all_ions_idx_for_nl": np.uint64,
        }

    def search(
            self,
            method="open",
            precursor_mz=None,
            peaks=None,
            ms2_tolerance_in_da=0.02,
            search_type=0,
            search_spectra_idx_min=0,
            search_spectra_idx_max=0,
            reverse=False
    ):
        """
        Perform identity-, open- or neutral loss search on the MS/MS spectra library.

        :param method:  The search method, can be "open" or "neutral_loss".
                        Set it to "open" for identity search and open search, set it to "neutral_loss" for neutral loss search.
        :param precursor_mz:    The precursor m/z of the query MS/MS spectrum, required for neutral loss search.
        :param peaks:   The peaks of the query MS/MS spectrum. The peaks need to be precleaned by "clean_spectrum" function.
        :param ms2_tolerance_in_da: The MS2 tolerance used when searching the MS/MS spectra, in Dalton. Default is 0.02.
        :param search_type: The search type, can be 0, 1 or 2.
                            Set it to 0 for searching the whole MS/MS spectra library.
                            Set it to 1 for searching a range of the MS/MS spectra library,
        :param search_spectra_idx_min:  The minimum index of the MS/MS spectra to search, required when search_type is 1.
        :param search_spectra_idx_max:  The maximum index of the MS/MS spectra to search, required when search_type is 1.
        """
        global library_mz
        if not self.index:
            return np.zeros(0, dtype=np.float32)
        if len(peaks) == 0:
            return (np.zeros(self.total_spectra_num, dtype=np.float32),
                    np.zeros(self.total_spectra_num, dtype=np.uint32),
                    np.zeros(self.total_spectra_num, dtype=np.float32),
                    np.zeros(self.total_spectra_num, dtype=np.float32))

        # Check peaks
        assert ms2_tolerance_in_da <= self.max_ms2_tolerance_in_da, "The MS2 tolerance is larger than the maximum MS2 tolerance."
        # assert abs(np.sum(np.square(peaks[:, 1])) - 1) < 1e-4, "The peaks are not normalized to sum to 1."
        assert (
                peaks.shape[0] <= 1 or np.min(peaks[1:, 0] - peaks[:-1, 0]) > self.max_ms2_tolerance_in_da * 2
        ), "The peaks array should be sorted by m/z, and the m/z difference between two adjacent peaks should be larger than 2 * max_ms2_tolerance_in_da."
        (
            all_ions_mz_idx_start,
            all_ions_mz,
            all_ions_intensity,
            all_ions_spec_idx,
            all_nl_mass_idx_start,
            all_nl_mass,
            all_nl_intensity,
            all_nl_spec_idx,
            all_ions_idx_for_nl,
        ) = self.index
        index_number_in_one_da = int(1 / self.mz_index_step)

        # Prepare the query spectrum
        peaks = self._preprocess_peaks(peaks)

        # Prepare the library
        if method == "open":
            library_mz_idx_start = all_ions_mz_idx_start
            library_mz = all_ions_mz
            library_peaks_intensity = all_ions_intensity
            library_spec_idx = all_ions_spec_idx
        else:  # neutral loss
            library_mz_idx_start = all_nl_mass_idx_start
            library_mz = all_nl_mass
            library_peaks_intensity = all_nl_intensity
            library_spec_idx = all_nl_spec_idx
            peaks[:, 0] = precursor_mz - peaks[:, 0]

        # Start searching
        # Calculate the similarity for this matched peak
        similarity_arr = np.zeros(self.total_spectra_num, dtype=np.float32)
        matched_cnt_arr = np.zeros(self.total_spectra_num, dtype=np.uint32)
        spec_usage_arr = np.zeros(self.total_spectra_num, dtype=np.float32)  # sum(matched peaks intensity) / sum(query peaks intensity)

        # a large empty 2D array to record matched peak pairs in query spectrum
        match_table_q = np.zeros((peaks.shape[0], self.total_spectra_num), dtype=np.float32)
        # a large empty 2D array to record matched peak pairs in ref. spectrum
        match_table_r = np.zeros((peaks.shape[0], self.total_spectra_num), dtype=np.float32)

        # loop through all peaks in query spectrum
        for k, (mz_query, intensity_query) in enumerate(peaks):
            # Determine the mz index range
            product_mz_idx_min = self._find_location_from_array_with_index(
                mz_query - ms2_tolerance_in_da, library_mz, library_mz_idx_start, "left", index_number_in_one_da
            )
            product_mz_idx_max = self._find_location_from_array_with_index(
                mz_query + ms2_tolerance_in_da, library_mz, library_mz_idx_start, "right", index_number_in_one_da
            )

            # intensity of matched peaks
            intensity_library = library_peaks_intensity[product_mz_idx_min:product_mz_idx_max]
            # library IDs of matched peaks
            modified_idx = library_spec_idx[product_mz_idx_min:product_mz_idx_max]

            # fill in the match table
            match_table_q[k, modified_idx] = intensity_query
            match_table_r[k, modified_idx] = intensity_library

        # search type 0: search the whole library
        if search_type == 0:
            # calculate spectral usage
            spec_usage_arr = np.sum(match_table_q, axis=0) / np.sum(peaks[:, 1])

            # transform query spectrum intensity in each column
            match_table_q = np.power(match_table_q, self.peak_intensity_power)
            peaks[:, 1] = np.power(peaks[:, 1], self.peak_intensity_power)

            if reverse:
                match_table_q = match_table_q / np.sqrt(np.sum(np.square(match_table_q), axis=0))
                # calculate similarity
                similarity_arr = np.sum(match_table_q * match_table_r, axis=0)

            else:
                match_table_q = match_table_q / np.sqrt(np.sum(np.square(peaks[:, 1]), axis=0))
                # calculate similarity
                similarity_arr = np.sum(match_table_q * match_table_r, axis=0)

            # count matched peaks
            matched_cnt_arr = np.sum(match_table_q > 0, axis=0)
        elif search_type == 1:
            # search type 1: search a range of the library, for identity search
            match_table_q = match_table_q[:, search_spectra_idx_min:search_spectra_idx_max]
            match_table_r = match_table_r[:, search_spectra_idx_min:search_spectra_idx_max]
            # calculate spectral usage
            part_spectral_usage = np.sum(match_table_q, axis=0) / np.sum(peaks[:, 1])

            # normalize query spectrum intensity in each column
            match_table_q = np.power(match_table_q, self.peak_intensity_power)
            peaks[:, 1] = np.power(peaks[:, 1], self.peak_intensity_power)

            if reverse:
                match_table_q_rev = match_table_q / np.sqrt(np.sum(np.square(match_table_q), axis=0))
                # calculate similarity
                part_similarity = np.sum(match_table_q_rev * match_table_r, axis=0)

            else:
                match_table_q = match_table_q / np.sqrt(np.sum(np.square(peaks[:, 1])))
                # calculate similarity
                part_similarity = np.sum(match_table_q * match_table_r, axis=0)

            # count matched peaks
            part_matched_cnt = np.sum(match_table_q > 0, axis=0)

            # update the similarity and matched count
            similarity_arr[search_spectra_idx_min:search_spectra_idx_max] = part_similarity
            matched_cnt_arr[search_spectra_idx_min:search_spectra_idx_max] = part_matched_cnt
            spec_usage_arr[search_spectra_idx_min:search_spectra_idx_max] = part_spectral_usage

        del match_table_q, match_table_r

        return similarity_arr, matched_cnt_arr, spec_usage_arr

    def search_hybrid(self, precursor_mz=None, peaks=None, ms2_tolerance_in_da=0.02, reverse=False):
        """
        Perform the hybrid search for the MS/MS spectra.

        :param precursor_mz: The precursor m/z of the MS/MS spectra.
        :param peaks: The peaks of the MS/MS spectra, needs to be cleaned with the "clean_spectrum" function.
        :param ms2_tolerance_in_da: The MS/MS tolerance in Da.
        """
        if not self.index:
            return np.zeros(0, dtype=np.float32)
        if len(peaks) == 0:
            return (np.zeros(self.total_spectra_num, dtype=np.float32),
                    np.zeros(self.total_spectra_num, dtype=np.uint32),
                    np.zeros(self.total_spectra_num, dtype=np.float32),
                    np.zeros(self.total_spectra_num, dtype=np.float32))

        # Check peaks
        assert ms2_tolerance_in_da <= self.max_ms2_tolerance_in_da, "The MS2 tolerance is larger than the maximum MS2 tolerance."
        # assert abs(np.sum(np.square(peaks[:, 1])) - 1) < 1e-4, "The peaks are not normalized to sum to 1."
        assert (
                peaks.shape[0] <= 1 or np.min(peaks[1:, 0] - peaks[:-1, 0]) > self.max_ms2_tolerance_in_da * 2
        ), "The peaks array should be sorted by m/z, and the m/z difference between two adjacent peaks should be larger than 2 * max_ms2_tolerance_in_da."
        (
            all_ions_mz_idx_start,
            all_ions_mz,
            all_ions_intensity,
            all_ions_spec_idx,
            all_nl_mass_idx_start,
            all_nl_mass,
            all_nl_intensity,
            all_nl_spec_idx,
            all_ions_idx_for_nl,
        ) = self.index
        index_number_in_one_da = int(1 / self.mz_index_step)

        # Prepare the query spectrum
        peaks = self._preprocess_peaks(peaks)

        # Go through all peak in the spectrum and determine the mz index range
        product_peak_match_idx_min = np.zeros(peaks.shape[0], dtype=np.uint64)
        product_peak_match_idx_max = np.zeros(peaks.shape[0], dtype=np.uint64)
        for peak_idx, (mz_query, _) in enumerate(peaks):
            # Determine the mz index range
            product_mz_idx_min = self._find_location_from_array_with_index(
                mz_query - ms2_tolerance_in_da, all_ions_mz, all_ions_mz_idx_start, "left", index_number_in_one_da
            )
            product_mz_idx_max = self._find_location_from_array_with_index(
                mz_query + ms2_tolerance_in_da, all_ions_mz, all_ions_mz_idx_start, "right", index_number_in_one_da
            )

            product_peak_match_idx_min[peak_idx] = product_mz_idx_min
            product_peak_match_idx_max[peak_idx] = product_mz_idx_max

        # a large empty 2D array to record matched peak pairs in query spectrum
        match_table_q = np.zeros((peaks.shape[0], self.total_spectra_num), dtype=np.float32)
        match_table_q_nl = np.zeros((peaks.shape[0], self.total_spectra_num), dtype=np.float32)  # for neutral loss
        # a large empty 2D array to record matched peak pairs in ref. spectrum
        match_table_r = np.zeros((peaks.shape[0], self.total_spectra_num), dtype=np.float32)
        match_table_r_nl = np.zeros((peaks.shape[0], self.total_spectra_num), dtype=np.float32)  # for neutral loss

        # Go through all the peaks in the spectrum and calculate the similarity
        for k, (mz, intensity) in enumerate(peaks):
            ###############################################################
            # Match the original product ion
            product_mz_idx_min = product_peak_match_idx_min[k]
            product_mz_idx_max = product_peak_match_idx_max[k]

            # intensity of matched peaks
            intensity_library = all_ions_intensity[product_mz_idx_min:product_mz_idx_max]
            # library IDs of matched peaks
            modified_idx = all_ions_spec_idx[product_mz_idx_min:product_mz_idx_max]

            # fill in the match table
            match_table_q[k, modified_idx] = intensity
            match_table_r[k, modified_idx] = intensity_library

            ###############################################################
            # Match the neutral loss ions
            mz_nl = precursor_mz - mz
            # Determine the mz index range
            neutral_loss_mz_idx_min = self._find_location_from_array_with_index(
                mz_nl - ms2_tolerance_in_da, all_nl_mass, all_nl_mass_idx_start, "left", index_number_in_one_da
            )
            neutral_loss_mz_idx_max = self._find_location_from_array_with_index(
                mz_nl + ms2_tolerance_in_da, all_nl_mass, all_nl_mass_idx_start, "right", index_number_in_one_da
            )

            # intensity of matched peaks
            intensity_library_nl = all_nl_intensity[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max].copy()
            modified_idx_nl = all_nl_spec_idx[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max]

            # Check if the neutral loss ion is already matched to other query peak as a product ion
            nl_matched_product_ion_idx = all_ions_idx_for_nl[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max]
            s1 = np.searchsorted(product_peak_match_idx_min, nl_matched_product_ion_idx, side="right")
            s2 = np.searchsorted(product_peak_match_idx_max - 1, nl_matched_product_ion_idx, side="left")
            intensity_library_nl[s1 > s2] = 0

            # # Check if this query peak is already matched to a product ion in the same library spectrum
            # duplicate_idx_in_nl = self._remove_duplicate_with_cpu(modified_idx, modified_idx_nl,
            #                                                       self.total_spectra_num)
            # intensity_library_nl[duplicate_idx_in_nl] = 0

            # fill in the match table for neutral loss
            match_table_q_nl[k, modified_idx_nl] = intensity
            match_table_r_nl[k, modified_idx_nl] = intensity_library_nl

        # merge the match table
        match_table_q_all = np.maximum(match_table_q, match_table_q_nl)
        match_table_r_all = np.maximum(match_table_r, match_table_r_nl)

        del match_table_q, match_table_r, match_table_q_nl, match_table_r_nl

        # calculate spectral usage
        spec_usage_arr = np.sum(match_table_q_all, axis=0) / np.sum(peaks[:, 1])

        # normalize query spectrum intensity in each column
        match_table_q_all = np.power(match_table_q_all, self.peak_intensity_power)
        peaks[:, 1] = np.power(peaks[:, 1], self.peak_intensity_power)

        if reverse:
            match_table_q_all_rev = match_table_q_all / np.sqrt(np.sum(np.square(match_table_q_all), axis=0))
            # calculate similarity
            similarity_arr = np.sum(match_table_q_all_rev * match_table_r_all, axis=0)

        else:
            match_table_q_all = match_table_q_all / np.sqrt(np.sum(np.square(peaks[:, 1]), axis=0))
            # calculate similarity
            similarity_arr = np.sum(match_table_q_all * match_table_r_all, axis=0)

        # count matched peaks
        matched_cnt_arr = np.sum(match_table_q_all > 0, axis=0)

        return similarity_arr, matched_cnt_arr, spec_usage_arr

    def _remove_duplicate_with_cpu(self, array_1, array_2, max_element):
        if len(array_1) + len(array_2) < 4_000_000:
            # When len(array_1) + len(array_2) < 4_000_000, this method is faster than array method
            aux = np.concatenate((array_1, array_2))
            aux_sort_indices = np.argsort(aux, kind="mergesort")
            aux = aux[aux_sort_indices]
            mask = aux[1:] == aux[:-1]
            array2_indices = aux_sort_indices[1:][mask] - array_1.size
            return array2_indices
        else:
            # When len(array_1) + len(array_2) > 4_000_000, this method is faster than sort method
            note = np.zeros(max_element, dtype=np.int8)
            note[array_1] = 1
            duplicate_idx = np.where(note[array_2] == 1)[0]
            return duplicate_idx

    def build_index(self, all_spectra_list: list, max_indexed_mz: float = 1500.00005):
        """
        Build the index for the MS/MS spectra library.

        The spectra provided to this function should be a dictionary in the format of {"precursor_mz": precursor_mz, "peaks": peaks}.
        The precursor_mz is the precursor m/z value of the MS/MS spectrum;
        The peaks is a numpy array which has been processed by the function "clean_spectrum".

        :param all_spectra_list:    A list of dictionaries in the format of {"precursor_mz": precursor_mz, "peaks": peaks},
                                    the spectra in the list need to be sorted by the precursor m/z.
        :param max_indexed_mz: The maximum m/z value that will be indexed. Default is 1500.00005.
        """

        # Get the total number of spectra and peaks
        total_peaks_num = np.sum([spectrum["peaks"].shape[0] for spectrum in all_spectra_list])
        total_spectra_num = len(all_spectra_list)
        # total_spectra_num can not be bigger than 2^32-1 (uint32), total_peak_num can not be bigger than 2^63-1 (int64)
        assert total_spectra_num < 2 ** 32 - 1, "The total spectra number is too big."
        assert total_peaks_num < 2 ** 63 - 1, "The total peaks number is too big."
        self.total_spectra_num = total_spectra_num
        self.total_peaks_num = total_peaks_num

        ############## Step 1: Collect the precursor m/z and peaks information. ##############
        dtype_peak_data = np.dtype(
            [
                ("ion_mz", np.float32),  # The m/z of the fragment ion.
                ("nl_mass", np.float32),  # The neutral loss mass of the fragment ion.
                ("intensity", np.float32),  # The intensity of the fragment ion.
                ("spec_idx", np.uint32),  # The index of the MS/MS spectra.
                ("peak_idx", np.uint64),
            ],
            align=True,
        )  # The index of the fragment ion.

        # Initialize the peak data array.
        peak_data = np.zeros(total_peaks_num, dtype=dtype_peak_data)
        peak_idx = 0

        print('Number of spectra:', len(all_spectra_list))

        # Adding the precursor m/z and peaks information to the peak data array.
        for idx, spectrum in enumerate(all_spectra_list):
            precursor_mz, peaks = spectrum["precursor_mz"], spectrum["peaks"]

            # Check the peaks array.
            assert peaks.ndim == 2, "The peaks array should be a 2D numpy array."
            assert peaks.shape[1] == 2, "The peaks array should be a 2D numpy array with the shape of [n, 2]."
            assert peaks.shape[0] > 0, "The peaks array should not be empty."
            assert abs(np.sum(np.square(peaks[:, 1])) - 1) < 1e-4, "The peaks array should be normalized to sum to 1."
            assert (
                    peaks.shape[0] <= 1 or np.min(peaks[1:, 0] - peaks[:-1, 0]) > self.max_ms2_tolerance_in_da * 2
            ), "The peaks array should be sorted by m/z, and the m/z difference between two adjacent peaks should be larger than 2 * max_ms2_tolerance_in_da."

            # Preprocess the peaks array.
            peaks = self._preprocess_peaks(peaks)

            # Assign the product ion m/z
            peak_data_item = peak_data[peak_idx: (peak_idx + peaks.shape[0])]
            peak_data_item["ion_mz"] = peaks[:, 0]
            # Assign the neutral loss mass
            peak_data_item["nl_mass"] = precursor_mz - peaks[:, 0]
            # Assign the intensity
            peak_data_item["intensity"] = peaks[:, 1]
            # Assign the spectrum index
            peak_data_item["spec_idx"] = idx
            # Set the peak index
            peak_idx += peaks.shape[0]

        ############## Step 2: Build the index by sort with product ions. ##############
        self.index = self._generate_index_from_peak_data(peak_data, max_indexed_mz)
        return self.index

    def _generate_index_from_peak_data(self, peak_data, max_indexed_mz):
        # Sort with precursor m/z.
        peak_data.sort(order="ion_mz")

        # Record the m/z, intensity, and spectrum index information for product ions.
        all_ions_mz = np.copy(peak_data["ion_mz"])
        all_ions_intensity = np.copy(peak_data["intensity"])
        all_ions_spec_idx = np.copy(peak_data["spec_idx"])

        # Assign the index of the product ions.
        peak_data["peak_idx"] = np.arange(0, self.total_peaks_num, dtype=np.uint64)

        # Build index for fast access to the ion's m/z.
        max_mz = min(np.max(all_ions_mz), max_indexed_mz)
        search_array = np.arange(0.0, max_mz, self.mz_index_step)
        all_ions_mz_idx_start = np.searchsorted(all_ions_mz, search_array, side="left").astype(np.int64)

        ############## Step 3: Build the index by sort with neutral loss mass. ##############
        # Sort with the neutral loss mass.
        peak_data.sort(order="nl_mass")

        # Record the m/z, intensity, spectrum index, and product ions index information for neutral loss ions.
        all_nl_mass = peak_data["nl_mass"]
        all_nl_intensity = peak_data["intensity"]
        all_nl_spec_idx = peak_data["spec_idx"]
        all_ions_idx_for_nl = peak_data["peak_idx"]

        # Build the index for fast access to the neutral loss mass.
        max_mz = min(np.max(all_nl_mass), max_indexed_mz)
        search_array = np.arange(0.0, max_mz, self.mz_index_step)
        all_nl_mass_idx_start = np.searchsorted(all_nl_mass, search_array, side="left").astype(np.int64)

        ############## Step 4: Save the index. ##############
        index = [
            all_ions_mz_idx_start,
            all_ions_mz,
            all_ions_intensity,
            all_ions_spec_idx,
            all_nl_mass_idx_start,
            all_nl_mass,
            all_nl_intensity,
            all_nl_spec_idx,
            all_ions_idx_for_nl,
        ]
        return index

    def _preprocess_peaks(self, peaks):
        """
        Preprocess the peaks.
        """
        peaks_clean = np.asarray(peaks).copy()
        return peaks_clean

    def _find_location_from_array_with_index(self, wanted_mz, mz_array, mz_idx_start_array, side, index_number):
        mz_min_int = (np.floor(wanted_mz * index_number)).astype(int)
        mz_max_int = mz_min_int + 1

        if mz_min_int >= len(mz_idx_start_array):
            mz_idx_search_start = mz_idx_start_array[-1]
        else:
            mz_idx_search_start = mz_idx_start_array[mz_min_int].astype(int)

        if mz_max_int >= len(mz_idx_start_array):
            mz_idx_search_end = len(mz_array)
        else:
            mz_idx_search_end = mz_idx_start_array[mz_max_int].astype(int) + 1

        return mz_idx_search_start + np.searchsorted(mz_array[mz_idx_search_start:mz_idx_search_end], wanted_mz,
                                                     side=side)

    def save_memory_for_multiprocessing(self):
        """
        Move the numpy array in the index to shared memory in order to save memory.
        This function is not required when you only use one thread to search the MS/MS spectra.
        When use multiple threads, this function is also not required but highly recommended, as it avoids the memory copy and saves a lot of memory and time.
        """
        if self._init_for_multiprocessing:
            return

        for i, array in enumerate(self.index):
            self.index[i] = _convert_numpy_array_to_shared_memory(array)
        self._init_for_multiprocessing = True

    def read(self, path_data=None):
        """
        Read the index from the specified path.
        """
        try:
            if path_data is None:
                path_data = self.path_data

            path_data = Path(path_data)
            self.index = []
            for name in self.index_names:
                self.index.append(np.fromfile(path_data / f"{name}.npy", dtype=self.index_dtypes[name]))

            with open(path_data / "information.json", "r") as f:
                information = json.load(f)
            self.mz_index_step = information["mz_index_step"]
            self.total_spectra_num = information["total_spectra_num"]
            self.total_peaks_num = information["total_peaks_num"]
            self.max_ms2_tolerance_in_da = information["max_ms2_tolerance_in_da"]
            return True
        except:
            return False

    def write(self, path_data=None):
        """
        Write the index to the specified path.
        """
        if path_data is None:
            path_data = self.path_data

        path_data = Path(path_data)
        path_data.mkdir(parents=True, exist_ok=True)
        for i, name in enumerate(self.index_names):
            self.index[i].tofile(str(path_data / f"{name}.npy"))
        information = {
            "mz_index_step": float(self.mz_index_step),
            "total_spectra_num": int(self.total_spectra_num),
            "total_peaks_num": int(self.total_peaks_num),
            "max_ms2_tolerance_in_da": float(self.max_ms2_tolerance_in_da),
        }
        with open(path_data / "information.json", "w") as f:
            json.dump(information, f)


def _convert_numpy_array_to_shared_memory(np_array, array_c_type=None):
    """
    The char table of shared memory can be find at:
    https://docs.python.org/3/library/struct.html#format-characters
    https://docs.python.org/3/library/array.html#module-array (This one is wrong!)
    The documentation of numpy.frombuffer can be find at:
    https://docs.scipy.org/doc/numpy/reference/generated/numpy.frombuffer.html
    Note: the char table is different from the char table in numpy
    """
    dim = np_array.shape
    num = reduce(lambda x, y: x * y, dim)
    if array_c_type is None:
        array_c_type = np_array.dtype.char
    base = multiprocessing.Array(array_c_type, num, lock=False)
    np_array_new = np.frombuffer(base, dtype=np_array.dtype).reshape(dim)
    np_array_new[:] = np_array
    return np_array_new


def _read_data_from_file(file_data, item_start, item_end):
    array_type = file_data.data_type
    type_size = array_type.itemsize
    file_data.seek(int(item_start * type_size))
    data = file_data.read(int(type_size * (item_end - item_start)))
    array = np.frombuffer(data, dtype=array_type)
    return array


class FlashCos:
    def __init__(self, max_ms2_tolerance_in_da=0.024, mz_index_step=0.0001, path_data=None,
                 peak_scale_k=8.0,
                 peak_intensity_power=1.0):
        self.precursor_mz_array = np.zeros(0, dtype=np.float32)
        self.peak_scale_k = peak_scale_k
        self.peak_intensity_power = peak_intensity_power

        self.similarity_search = FlashCosCore(path_data=path_data,
                                              max_ms2_tolerance_in_da=max_ms2_tolerance_in_da,
                                              mz_index_step=mz_index_step,
                                              peak_scale_k=peak_scale_k,
                                              peak_intensity_power=peak_intensity_power)

    def identity_search(self, precursor_mz, peaks, ms1_tolerance_in_da, ms2_tolerance_in_da,
                        reverse, **kwargs):
        """
        Run the identity search, the query spectrum should be preprocessed.

        For super large spectral library, directly identity search is not recommended. To do the identity search on super large spectral library,
        divide the spectral library into several parts, build the index for each part, and then do the identity search on each part will be much faster.

        :param precursor_mz:    The precursor m/z of the query spectrum.
        :param peaks:           The peaks of the query spectrum, should be the output of `clean_spectrum()` function.
        :param ms1_tolerance_in_da:  The MS1 tolerance in Da.
        :param ms2_tolerance_in_da:  The MS2 tolerance in Da.
        :return:    The similarity score for each spectrum in the library, a numpy array with shape (N,), N is the number of spectra in the library.
        """
        precursor_mz_min = precursor_mz - ms1_tolerance_in_da
        precursor_mz_max = precursor_mz + ms1_tolerance_in_da
        spectra_idx_min = np.searchsorted(self.precursor_mz_array, precursor_mz_min, side='left')
        spectra_idx_max = np.searchsorted(self.precursor_mz_array, precursor_mz_max, side='right')
        if spectra_idx_min >= spectra_idx_max:
            return (np.zeros(self.similarity_search.total_spectra_num, dtype=np.float32),
                    np.zeros(self.similarity_search.total_spectra_num, dtype=np.float32),
                    np.zeros(self.similarity_search.total_spectra_num, dtype=np.float32),
                    np.zeros(self.similarity_search.total_spectra_num, dtype=np.float32))
        else:
            return self.similarity_search.search(method="open", peaks=peaks, precursor_mz=precursor_mz,
                                                 ms2_tolerance_in_da=ms2_tolerance_in_da,
                                                 search_type=1, search_spectra_idx_min=spectra_idx_min,
                                                 search_spectra_idx_max=spectra_idx_max,
                                                 reverse=reverse)

    def open_search(self, precursor_mz, peaks, ms2_tolerance_in_da, reverse, **kwargs):
        """
        Run the open search, the query spectrum should be preprocessed.

        :param peaks:           The peaks of the query spectrum, should be the output of `clean_spectrum()` function.
        :param ms2_tolerance_in_da:  The MS2 tolerance in Da.
        :return:    The similarity score for each spectrum in the library, a numpy array with shape (N,), N is the number of spectra in the library.
        """
        return self.similarity_search.search(method="open", peaks=peaks, precursor_mz=precursor_mz,
                                             ms2_tolerance_in_da=ms2_tolerance_in_da, search_type=0,
                                             reverse=reverse)

    def neutral_loss_search(self, precursor_mz, peaks, ms2_tolerance_in_da, reverse, **kwargs):
        """
        Run the neutral loss search, the query spectrum should be preprocessed.

        :param precursor_mz:    The precursor m/z of the query spectrum.
        :param peaks:           The peaks of the query spectrum, should be the output of `clean_spectrum()` function.
        :param ms2_tolerance_in_da:  The MS2 tolerance in Da.
        :return:    The similarity score for each spectrum in the library, a numpy array with shape (N,), N is the number of spectra in the library.
        """
        return self.similarity_search.search(method="neutral_loss", precursor_mz=precursor_mz, peaks=peaks,
                                             ms2_tolerance_in_da=ms2_tolerance_in_da, search_type=0,
                                             reverse=reverse)

    def hybrid_search(self, precursor_mz, peaks, ms2_tolerance_in_da, reverse, **kwargs):
        """
        Run the hybrid search, the query spectrum should be preprocessed.

        :param precursor_mz:    The precursor m/z of the query spectrum.
        :param peaks:           The peaks of the query spectrum, should be the output of `clean_spectrum()` function.
        :param ms2_tolerance_in_da:  The MS2 tolerance in Da.

        :return:    The similarity score for each spectrum in the library, a numpy array with shape (N,), N is the number of spectra in the library.
        """
        return self.similarity_search.search_hybrid(precursor_mz=precursor_mz, peaks=peaks,
                                                    ms2_tolerance_in_da=ms2_tolerance_in_da,
                                                    reverse=reverse)

    def clean_spectrum_for_search(self,
                                  precursor_mz,
                                  peaks,
                                  precursor_ions_removal_da: float = 1.6,
                                  noise_threshold=0.01,
                                  min_ms2_difference_in_da: float = 0.05,
                                  peak_scale_k: float = 8.0,
                                  peak_intensity_power: float = 1.0):
        """
        Clean the MS/MS spectrum, need to be called before any search.

        :param precursor_mz:    The precursor m/z of the spectrum.
        :param peaks:           The peaks of the spectrum, should be a list or numpy array with shape (N, 2), N is the number of peaks. The format of the peaks is [[mz1, intensity1], [mz2, intensity2], ...].
        :param precursor_ions_removal_da:   The ions with m/z larger than precursor_mz - precursor_ions_removal_da will be removed.
                                            Default is 1.6.
        :param noise_threshold: The intensity threshold for removing the noise peaks. The peaks with intensity smaller than noise_threshold * max(intensity)
                                will be removed. Default is 0.01.
        :param min_ms2_difference_in_da:    The minimum difference between two peaks in the MS/MS spectrum. Default is 0.05.
        :param peak_intensity_power:    The power of the peak intensity. Default is 1.0.
        """

        if precursor_ions_removal_da is not None:
            max_mz = precursor_mz - precursor_ions_removal_da
        else:
            max_mz = None

        return preprocess_ms2(peaks=peaks,
                              prec_mz=precursor_mz,
                              min_mz=-1,
                              max_mz=max_mz,
                              relative_intensity_cutoff=noise_threshold,
                              min_ms2_difference_in_da=min_ms2_difference_in_da,
                              min_ms2_difference_in_ppm=-1,
                              top6_every_50da=True,
                              peak_scale_k=peak_scale_k,
                              peak_intensity_power=peak_intensity_power,
                              peak_norm='sum_sq')

    def search(self,
               precursor_mz, peaks,
               ms1_tolerance_in_da=0.01,
               ms2_tolerance_in_da=0.02,
               method: Union[str, List] = "all",  # identity, open, neutral_loss, hybrid, all
               precursor_ions_removal_da: Union[float, None] = 1.6,
               noise_threshold=0.01,
               min_ms2_difference_in_da: float = 0.05,
               reverse: bool = True):
        """
        Run the Flash search for the query spectrum.
        :return:    A dictionary with the search results. The keys are "identity_search", "open_search", "neutral_loss_search", "hybrid_search", and the values are the search results for each method.
        """

        if method == "all":
            method = {"identity", "open", "neutral_loss", "hybrid"}
        elif isinstance(method, str):
            method = {method}

        # do not normalize the intensity here, as the intensity will be normalized after alignment
        peaks = preprocess_ms2(peaks=peaks,
                               prec_mz=precursor_mz,
                               min_mz=-1,
                               max_mz=precursor_mz - precursor_ions_removal_da,
                               relative_intensity_cutoff=noise_threshold,
                               min_ms2_difference_in_da=min_ms2_difference_in_da,
                               min_ms2_difference_in_ppm=-1,
                               top6_every_50da=False,
                               peak_intensity_power=1.0,
                               peak_norm=None)

        result = {}
        if "identity" in method:
            temp_result = self.identity_search(precursor_mz=precursor_mz,
                                               peaks=peaks,
                                               ms1_tolerance_in_da=ms1_tolerance_in_da,
                                               ms2_tolerance_in_da=ms2_tolerance_in_da,
                                               reverse=reverse)
            result["identity_search"] = _clean_search_result(temp_result)
        if "open" in method:
            temp_result = self.open_search(precursor_mz=precursor_mz,
                                           peaks=peaks,
                                           ms2_tolerance_in_da=ms2_tolerance_in_da,
                                           reverse=reverse)
            result["open_search"] = _clean_search_result(temp_result)
        if "neutral_loss" in method:
            temp_result = self.neutral_loss_search(precursor_mz=precursor_mz,
                                                   peaks=peaks,
                                                   ms2_tolerance_in_da=ms2_tolerance_in_da,
                                                   reverse=reverse)
            result["neutral_loss_search"] = _clean_search_result(temp_result)
        if "hybrid" in method:
            temp_result = self.hybrid_search(precursor_mz=precursor_mz,
                                             peaks=peaks,
                                             ms2_tolerance_in_da=ms2_tolerance_in_da,
                                             reverse=reverse)
            result["hybrid_search"] = _clean_search_result(temp_result)
        return result

    def search_ref_premz_in_qry(self,
                                precursor_mz,
                                peaks,
                                mass_shift: float = 0.0,
                                ms2_tolerance_in_da=0.02,
                                method: str = "open"):
        """
        Search reference precursor m/z in query spectra. Return the matched peak intensity in the query spectra.

        :param precursor_mz: precursor m/z of the query spectra
        :param peaks: peaks of the query spectra (preprocessed)
        :param mass_shift: reference precursor m/z shift; search for the reference precursor m/z + mass_shift in query spectra
        :param ms2_tolerance_in_da: MS2 tolerance in Da
        :param method: open or neutral_loss
        :return:  A numpy array with shape (N,), N is the number of spectra in the library.
        """
        # arr to return
        ref_intensity_arr = np.zeros(len(self.precursor_mz_array), dtype=np.float32)

        if len(peaks) > 0:
            # shift the precursor m/z
            target_mass_arr = self.precursor_mz_array + mass_shift

            peaks[:, 1] = peaks[:, 1] / max(peaks[:, 1])

            if method == "open":
                qry_mass_arr = peaks[:, 0]
            else:
                qry_mass_arr = precursor_mz - peaks[:, 0]

            # Compute the absolute difference matrix
            diff_matrix = np.abs(qry_mass_arr[:, np.newaxis] - target_mass_arr)

            # Find the minimum difference for each target mass
            min_diffs = np.min(diff_matrix, axis=0)

            # Find the index of the minimum difference for each target mass
            min_diff_idx = np.argmin(diff_matrix, axis=0)

            # Mask where the minimum difference exceeds the tolerance
            valid_matches = min_diffs <= ms2_tolerance_in_da
            matched_indices = min_diff_idx[valid_matches]

            # Assign intensities for valid matches
            ref_intensity_arr[valid_matches] = peaks[matched_indices, 1]

        return ref_intensity_arr

    def build_index(self,
                    all_spectra_list: list = None,
                    max_indexed_mz: float = 1500.00005,
                    precursor_ions_removal_da: Union[float, None] = 1.6,
                    noise_threshold=0.01,
                    min_ms2_difference_in_da: float = 0.05,
                    clean_spectra: bool = True):
        """
        Set the library spectra for search. This function will sort the spectra by the precursor m/z and output the sorted spectra list.

        :return:    If the all_spectra_list is provided, this function will return the sorted spectra list.
        """

        # Sort the spectra by the precursor m/z.
        all_sorted_spectra_list = sorted(all_spectra_list, key=lambda x: x["precursor_mz"])

        # Clean the spectra, and collect the non-empty spectra
        all_spectra_list = []
        all_metadata_list = []
        for spec in all_sorted_spectra_list:
            # Clean the peaks
            if clean_spectra:
                spec["peaks"] = self.clean_spectrum_for_search(peaks=spec["peaks"],
                                                               precursor_mz=spec["precursor_mz"],
                                                               precursor_ions_removal_da=precursor_ions_removal_da,
                                                               noise_threshold=noise_threshold,
                                                               min_ms2_difference_in_da=min_ms2_difference_in_da,
                                                               peak_scale_k=self.peak_scale_k,
                                                               peak_intensity_power=self.peak_intensity_power)

            if len(spec["peaks"]) > 0:
                all_spectra_list.append(spec)
                all_metadata_list.append(pickle.dumps(spec))

        # Extract precursor m/z array
        self.precursor_mz_array = np.array([spec["precursor_mz"] for spec in all_spectra_list], dtype=np.float32)

        # Extract metadata array
        all_metadata_len = np.array([0] + [len(metadata) for metadata in all_metadata_list], dtype=np.uint64)
        self.metadata_loc = np.cumsum(all_metadata_len).astype(np.uint64)
        self.metadata = np.frombuffer(b''.join(all_metadata_list), dtype=np.uint8)

        # Call father class to build the index.
        self.similarity_search.build_index(all_spectra_list, max_indexed_mz)
        return all_spectra_list

    def __getitem__(self, index):
        """
        Get the MS/MS metadata by the index.

        :param index:   The index of the MS/MS spectrum.
        :return:    The MS/MS spectrum in the format of {"precursor_mz": precursor_mz, "peaks": peaks}.
        """
        if self.metadata is not None:
            spectrum = pickle.loads(self.metadata[self.metadata_loc[index]:self.metadata_loc[index + 1]].tobytes())
        else:
            spectrum = {"precursor_mz": self.precursor_mz_array[index]}
        return spectrum

    def write(self, path_data=None):
        """
        Write the MS/MS spectral library to a file.
        """
        if path_data is None:
            path_data = self.similarity_search.path_data
        else:
            path_data = Path(path_data)

        path_data = Path(path_data)
        path_data.mkdir(parents=True, exist_ok=True)

        self.precursor_mz_array.tofile(str(path_data / "precursor_mz.npy"))
        self.metadata.tofile(str(path_data / "metadata.npy"))
        self.metadata_loc.tofile(str(path_data / "metadata_loc.npy"))

        self.similarity_search.write(path_data)

    def read(self, path_data=None):
        """
        Read the MS/MS spectral library from a file.
        """
        if path_data is None:
            path_data = self.similarity_search.path_data
        else:
            path_data = Path(path_data)

        self.precursor_mz_array = np.fromfile(str(path_data / "precursor_mz.npy"), dtype=np.float32)
        self.metadata = np.fromfile(str(path_data / "metadata.npy"), dtype=np.uint8)
        self.metadata_loc = np.fromfile(str(path_data / "metadata_loc.npy"), dtype=np.uint64)

        return self.similarity_search.read(path_data)

    def save_memory_for_multiprocessing(self):
        """
        Save the memory for multiprocessing. This function will move the numpy array in the index to shared memory in order to save memory.

        This function is not required when you only use one thread to search the MS/MS spectra.
        When use multiple threads, this function is also not required but highly recommended, as it avoids the memory copy and saves a lot of memory and time.

        :return:    None
        """
        self.similarity_search.save_memory_for_multiprocessing()


def _clean_search_result(temp_result):
    score_arr = np.where(np.isfinite(temp_result[0]), temp_result[0], 0)
    temp_result = (score_arr, temp_result[1], temp_result[2])
    return temp_result


# test
if __name__ == "__main__":
    # cosine between two vectors
    def scale(peaks, prec_mz):
        scaling_factor = peaks[:, 0] / prec_mz * 8.0
        # print('scaling_factor:', scaling_factor)
        return peaks[:, 1] * np.exp(scaling_factor)


    def cosine_similarity(v1, v2):
        return np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))

    # load spectral library
    spectral_library = [{
        "id": "Demo spectrum 1",
        "precursor_mz": 150.0,
        "peaks": [[100.0, 1.0], [101.0, 1.0], [103.0, 1.0]]
    }, {
        "id": "Demo spectrum 2",
        "precursor_mz": 200.0,
        "peaks": np.array([[100.0, 10], [101.0, 5], [102.0, 20]], dtype=np.float32),
        "metadata": "ABC"
    }, {
        "id": "Demo spectrum 3",
        "precursor_mz": 250.0,
        "peaks": np.array([[100.0, 100], [101.0, 0.5], [202.0, 20], [205.0, 20]], dtype=np.float32),
        "XXX": "YYY",
    }, {
        "precursor_mz": 350.0,
        "peaks": [[200.0, 1.0], [250.0, 2.0], [349.0, 1.0]]}]

    ms2_tol = 0.02
    # search
    search_eng = FlashCos(max_ms2_tolerance_in_da=ms2_tol * 1.05,
                          mz_index_step=0.0001,
                          peak_scale_k=8.0,
                          peak_intensity_power=0.5)
    search_eng.build_index(spectral_library,
                           max_indexed_mz=2000,
                           precursor_ions_removal_da=0.5,
                           noise_threshold=0.0,
                           min_ms2_difference_in_da=ms2_tol * 2.2,
                           clean_spectra=True)

    # test
    _precursor_mz = 250.0
    peaks = np.array([[100.0, 100], [101.0, 0.5], [202.0, 20], [203.0, 0], [204.0, 0], [205.0, 20]])

    print(cosine_similarity(np.array([100, 0.5, 20, 0, 0, 20]), np.array([100, 30, 4, 3, 10, 0])))
    # print(scale(peaks, _precursor_mz))
    print(cosine_similarity(scale(peaks, _precursor_mz), np.array([100, 30, 4, 3, 10, 0])))
    print(cosine_similarity(np.sqrt(scale(peaks, _precursor_mz)), np.sqrt(np.array([100, 30, 4, 3, 10, 0]))))

    print('reverse')
    _peaks = np.array([[100.0, 100], [101.0, 0.5], [202.0, 20], [205.0, 20]])
    # print(scale(_peaks, _precursor_mz))
    print(cosine_similarity(scale(_peaks, _precursor_mz), np.array([100, 30, 4, 0])))
    print(cosine_similarity(np.sqrt(scale(_peaks, _precursor_mz)), np.sqrt(np.array([100, 30, 4, 0]))))

    search_result = search_eng.search(
        precursor_mz=250.0,
        peaks=np.array([[100.0, 100], [101.0, 30], [202.0, 4.0], [203.0, 3.0], [204.0, 10], [205.0, 0]]),
        ms1_tolerance_in_da=0.02,
        ms2_tolerance_in_da=ms2_tol,
        method="identity",  # "identity", "open", "neutral_loss", "hybrid", "all", or list of the above
        precursor_ions_removal_da=0.5,
        noise_threshold=0.0,
        min_ms2_difference_in_da=ms2_tol * 2.2,
        reverse=True
    )

    # score_arr, matched_peak_arr, spec_usage_arr, scaled_score_arr = search_result['identity_search']

    print(search_result)



import pandas as pd


def find_demo(file):
    """
    find demo: one annotation, multi adduct forms
    """

    df = pd.read_csv(
        f'/Users/shipei/Documents/projects/ms1_id/data/nist/data_1/output_gnps/{file}',
        sep='\t', low_memory=False)

    # Filter out rows with null MS1_similarity
    df = df[df['MS1_similarity'].notnull()].reset_index(drop=True)

    # Group by MS1_inchikey and aggregate
    grouped = df.groupby('MS1_inchikey').agg({
        'MS1_precursor_type': lambda x: list(set(x)),  # List unique precursor types
        'MS1_annotation': 'first',  # Keep the first MS1_annotation for reference
        'MS1_similarity': 'max',  # Keep the max MS1_similarity for reference
    })

    # Add a column for the count of unique adduct forms
    grouped['adduct_count'] = grouped['MS1_precursor_type'].apply(len)

    # Filter groups with more than one unique MS1_precursor_type
    multi_adduct = grouped[grouped['adduct_count'] > 1]

    # Sort by number of unique adducts (descending) and then by MS1_similarity (descending)
    multi_adduct = multi_adduct.sort_values(['adduct_count', 'MS1_similarity'], ascending=[False, False])

    # Display the results
    print(f"Total unique InChIKeys with multiple adduct forms: {len(multi_adduct)}")
    print("\nTop 10 entries:")
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', None)
    print(multi_adduct.head(10))

    # save
    multi_adduct.to_csv('data/multi_adduct_forms.tsv', sep='\t', index=True)

    print("\nResults have been saved to 'multi_adduct_forms.csv'")


if __name__ == '__main__':
    find_demo('NIST_pool_1_10eV_feature_table.tsv')



import pandas as pd


def add_ion_mode(search):
    """
    only keep rows with ion mode matched
    """
    nist20 = pd.read_csv('/Users/shipei/Documents/projects/ms2/ms2_lib/nist20/nist20/nist20.tsv',
                         sep='\t', low_memory=False)
    nist20['Ion_mode'] = nist20['Ion_mode'].str.lower()
    # dict from NIST_No to ion mode
    nist_no_to_ion_mode = nist20.set_index('NIST_No')['Ion_mode'].to_dict()

    gnps = pd.read_pickle('/Users/shipei/Documents/projects/ms1_id/data/ms2db/gnps.pkl')
    gnps['ion_mode'] = gnps['ion_mode'].str[0].str.lower()
    gnps_id_to_ion_mode_dict = dict(zip(gnps['db_id'], gnps['ion_mode']))

    df = pd.read_csv(f'{search}_results_mces.tsv', sep='\t')

    df['matched_gnps_id'] = df['matched_db_id'].apply(lambda x: x.split('=')[1].split(';')[0])  # gnps
    df['matched_ion_mode'] = df['matched_gnps_id'].map(gnps_id_to_ion_mode_dict)

    df['qry_ion_mode'] = df['qry_db_id'].map(nist_no_to_ion_mode)

    df['ion_mode_match'] = df['qry_ion_mode'] == df['matched_ion_mode']

    # save
    df.to_csv(f'{search}_results_mces.tsv', sep='\t', index=False)
    print(df['ion_mode_match'].value_counts())


def main_fdr(search, score_cutoff=0.8, min_matched_peak=4, mces_dist_cutoff=2):
    df = pd.read_csv(f'{search}_results_mces.tsv', sep='\t')

    df = df[(pd.notnull(df['mces_dist'])) & (df['ion_mode_match'])].reset_index(drop=True)
    df = df[(df['score'] >= score_cutoff) & (df['matched_peak'] >= min_matched_peak)].reset_index(drop=True)

    df['match'] = df['mces_dist'] <= mces_dist_cutoff

    print('====================')
    print(df['match'].value_counts())
    print('FDR: ', sum(df['match'] == False) / df.shape[0])

    # for each qry_id, retain the best match
    df = df.sort_values('score', ascending=False).drop_duplicates('qry_db_id').reset_index(drop=True)
    print('====================')
    print('After best match selection:')
    print(df['match'].value_counts())
    print('FDR: ', sum(df['match'] == False) / df.shape[0])


def search_merged_fdr(score_cutoff=0.8, min_matched_peak=4, mces_dist_cutoff=2):
    mode = 'k10'
    df1 = pd.read_csv('k0_results_mces.tsv', sep='\t')
    df2 = pd.read_csv(f'{mode}_results_mces.tsv', sep='\t')

    df = pd.concat([df1, df2]).reset_index(drop=True)
    df = df[(pd.notnull(df['mces_dist'])) & (df['ion_mode_match'])].reset_index(drop=True)

    df = df[(df['score'] >= score_cutoff) & (df['matched_peak'] >= min_matched_peak)].reset_index(drop=True)
    df.drop_duplicates(subset=['qry_db_id', 'matched_db_id'], inplace=True)

    df['match'] = df['mces_dist'] <= mces_dist_cutoff

    print('====================')
    print(df['match'].value_counts())

    print('FDR: ', sum(df['match'] == False) / df.shape[0])

    # for each qry_id, retain the best match
    df = df.sort_values('score', ascending=False).drop_duplicates('qry_db_id').reset_index(drop=True)
    print('====================')
    print('After best match selection:')
    print(df['match'].value_counts())
    print('FDR: ', sum(df['match'] == False) / df.shape[0])


if __name__ == '__main__':
    # add_ion_mode('k0')
    # add_ion_mode('k10')

    mces_dist_cutoff = 4

    min_score = 0.7
    min_peaks = 3
    main_fdr('k0', min_score, min_peaks, mces_dist_cutoff)

    main_fdr('k10', min_score, min_peaks, mces_dist_cutoff)

    search_merged_fdr(min_score, min_peaks, mces_dist_cutoff)



import pandas as pd
import pickle


def _ms1_ms2_annotation_overlap(mode='hilic_pos'):
    ms1df = pd.read_csv(
        f'/Users/shipei/Documents/projects/ms1_id/data/PR000639/output/{mode}_aligned_feature_table.tsv',
        sep='\t', low_memory=False)

    if mode == 'hilic_pos':
        ms2df = pd.read_csv(
            '/data/MSV000087562/HILIC_pos_iHMPpool/aligned_feature_table.tsv',
            sep='\t', low_memory=False)
    elif mode == 'c18_neg':
        ms2df = pd.read_csv(
            '/data/MSV000087562/C18_neg_iHMPpool/aligned_feature_table.tsv',
            sep='\t', low_memory=False)

    ms1df = ms1df[ms1df['MS1_similarity'].notnull()].reset_index(drop=True)
    ms1_inchikey = ms1df['MS1_inchikey'].unique().tolist()
    # remove empty inchikey
    ms1_inchikey = [i for i in ms1_inchikey if i]

    ms2df = ms2df[ms2df['similarity'].notnull()].reset_index(drop=True)
    ms2_inchikey = ms2df['InChIKey'].unique().tolist()
    # remove empty inchikey
    ms2_inchikey = [i for i in ms2_inchikey if i]

    print(f'Number of MS1 inchikey: {len(ms1_inchikey)}')
    print(f'Number of MS2 inchikey: {len(ms2_inchikey)}')

    common_inchikey = set(ms1_inchikey).intersection(set(ms2_inchikey))
    print(f'Number of common inchikey: {len(common_inchikey)}')


def four_mode_annotation():
    df1 = pd.read_csv(
        '/data/PR000639/output/hilic_pos_aligned_feature_table.tsv',
        sep='\t', low_memory=False)
    df2 = pd.read_csv(
        '/data/PR000639/output/hilic_neg_aligned_feature_table.tsv',
        sep='\t', low_memory=False)
    df3 = pd.read_csv('/data/PR000639/output/c8_pos_aligned_feature_table.tsv',
                      sep='\t', low_memory=False)
    df4 = pd.read_csv('/data/PR000639/output/c18_neg_aligned_feature_table.tsv',
                      sep='\t', low_memory=False)

    print('Total features in each mode:')
    print(f'HILIC pos: {df1.shape[0]}')
    print(f'HILIC neg: {df2.shape[0]}')
    print(f'C8 pos: {df3.shape[0]}')
    print(f'C18 neg: {df4.shape[0]}')

    df1 = df1[df1['MS1_similarity'].notnull()].reset_index(drop=True)
    df2 = df2[df2['MS1_similarity'].notnull()].reset_index(drop=True)
    df3 = df3[df3['MS1_similarity'].notnull()].reset_index(drop=True)
    df4 = df4[df4['MS1_similarity'].notnull()].reset_index(drop=True)

    df1_inchikey = df1['MS1_inchikey'].unique().tolist()
    df2_inchikey = df2['MS1_inchikey'].unique().tolist()
    df3_inchikey = df3['MS1_inchikey'].unique().tolist()
    df4_inchikey = df4['MS1_inchikey'].unique().tolist()

    # remove nan inchikey
    df1_inchikey = [i for i in df1_inchikey if not pd.isnull(i)]
    df2_inchikey = [i for i in df2_inchikey if not pd.isnull(i)]
    df3_inchikey = [i for i in df3_inchikey if not pd.isnull(i)]
    df4_inchikey = [i for i in df4_inchikey if not pd.isnull(i)]

    print('Number of MS1 inchikey in each mode:')
    print(f'HILIC pos: {len(df1_inchikey)}')
    print(f'HILIC neg: {len(df2_inchikey)}')
    print(f'C8 pos: {len(df3_inchikey)}')
    print(f'C18 neg: {len(df4_inchikey)}')

    # total unique inchikey
    total_inchikey = set(df1_inchikey).union(set(df2_inchikey)).union(set(df3_inchikey)).union(set(df4_inchikey))
    print(f'Total unique inchikey: {len(total_inchikey)}')

    common_inchikey = set(df1_inchikey).intersection(set(df2_inchikey)).intersection(set(df3_inchikey)).intersection(
        set(df4_inchikey))
    print(f'Number of common inchikey: {len(common_inchikey)}')

    # save
    pickle.dump(df1_inchikey, open('hilic_pos_inchikey.pkl', 'wb'))
    pickle.dump(df2_inchikey, open('hilic_neg_inchikey.pkl', 'wb'))
    pickle.dump(df3_inchikey, open('c8_pos_inchikey.pkl', 'wb'))
    pickle.dump(df4_inchikey, open('c18_neg_inchikey.pkl', 'wb'))


def add_metadata():
    modes = ['hilic_pos', 'hilic_neg', 'c8_pos', 'c18_neg']

    metadata = pd.read_csv('/data/PR000639/metadata/metadata.csv')
    # dictionary from local_sample_id to Diagnosis and sex
    sample_id_to_diagnosis = dict(zip(metadata['local_sample_id'], metadata['Diagnosis']))
    sample_id_to_sex = dict(zip(metadata['local_sample_id'], metadata['sex']))

    for mode in modes:
        df = pd.read_csv(
            f'/Users/shipei/Documents/projects/ms1_id/data/PR000639/output/{mode}_aligned_feature_table.tsv',
            sep='\t', low_memory=False)

        meta_df = pd.read_csv(f'/Users/shipei/Documents/projects/ms1_id/data/PR000639/metadata/{mode}.csv')
        meta_df['raw_name'] = meta_df['Raw file name'].apply(lambda x: x.split('.')[0])
        # dictionary from raw_name to 'Sample ID'
        raw_name_to_sample_id = dict(zip(meta_df['raw_name'], meta_df['Sample ID']))

        # change df column names
        col_diagnosis = []
        col_sex = []
        new_columns = []
        for col in df.columns:
            if col in raw_name_to_sample_id.keys():
                sample_id = raw_name_to_sample_id[col]
                new_columns.append(sample_id)

                # pooled ref. mixture
                if sample_id.startswith('PREF'):
                    col_diagnosis.append(None)
                    col_sex.append(None)
                else:
                    col_diagnosis.append(sample_id_to_diagnosis[sample_id])
                    col_sex.append(sample_id_to_sex[sample_id])
            else:
                new_columns.append(col)
                col_diagnosis.append(None)
                col_sex.append(None)

        # Rename columns
        df.columns = new_columns

        # Add two rows for diagnosis and sex
        diagnosis_row = pd.Series(col_diagnosis, index=df.columns, name='diagnosis')
        sex_row = pd.Series(col_sex, index=df.columns, name='sex')

        # Concatenate the new rows with the existing DataFrame
        df = pd.concat([diagnosis_row.to_frame().T, sex_row.to_frame().T, df], axis=0)

        # Reset the index
        df = df.reset_index(drop=True)

        # Save the updated DataFrame
        df.to_csv(
            f'/Users/shipei/Documents/projects/ms1_id/data/PR000639/{mode}_aligned_feature_table_with_metadata.tsv',
            sep='\t', index=False)


if __name__ == '__main__':
    ################
    # ms2 data are targetedly collected
    # _ms1_ms2_annotation_overlap('hilic_pos')
    # _ms1_ms2_annotation_overlap('c18_neg')

    ################
    four_mode_annotation()
    '''
    Total features in each mode:
    HILIC pos: 20401
    HILIC neg: 5807
    C8 pos: 9502
    C18 neg: 9079
    Number of MS1 inchikey in each mode:
    HILIC pos: 540
    HILIC neg: 59
    C8 pos: 28
    C18 neg: 225
    Total unique inchikey: 794
    '''

    # add_metadata()



import matplotlib.pyplot as plt
from pyteomics import mzml
import numpy as np
from matplotlib.colors import LinearSegmentedColormap


def get_xics(mzml_file, target_mzs, ms_level=1, rt_range=None, mz_tol=0.02):
    """
    Function to extract multiple XICs for given masses, MS level, and retention time period.
    :param mzml_file: Path to the mzML file.
    :param target_mzs: List of target m/z values.
    :param ms_level: MS level to consider (1 or 2).
    :param rt_range: Retention time range as a tuple (start, end). If None, extract for all.
    :param mz_tol: Tolerance for the m/z value.
    :return: Dictionary of XICs and retention times.
    """
    xics = {target_mz: [] for target_mz in target_mzs}
    times = []

    with mzml.read(mzml_file) as reader:
        for spectrum in reader:
            if spectrum['ms level'] == ms_level:
                time = spectrum['scanList']['scan'][0]['scan start time']

                # Check if the retention time is within the specified range
                if rt_range is not None and not (rt_range[0] <= time <= rt_range[1]):
                    continue

                if len(times) == 0 or times[-1] != time:
                    times.append(time)

                mz_array = spectrum['m/z array']
                intensity_array = spectrum['intensity array']

                for target_mz in target_mzs:
                    xic_intensity = np.sum(
                        intensity_array[(mz_array >= target_mz - mz_tol) & (mz_array <= target_mz + mz_tol)])
                    xics[target_mz].append(xic_intensity)

    return times, xics


def plot_xics(times, xics, target_mzs, rt_range=None,
              fig_size=(3, 3), linewidth=2.5,
              y_max=None, x_axis=True,
              legend=True,
              save=False, name='xics.svg'):
    """
    Function to plot XICs.
    :param times: List of retention times.
    :param xics: Dictionary of XICs.
    :param target_mzs: List of target m/z values.
    :param rt_range: Retention time range as a tuple (start, end) in minutes.
    """
    colors = ['#56648a', '#6280a5', '#8ca5c0', '#8d7e95', '#facaa9', '#ca9a96']
    plt.figure(figsize=fig_size)

    if len(target_mzs) > len(colors):
        # Create a custom colormap
        cmap = LinearSegmentedColormap.from_list("custom", colors, N=len(target_mzs))
        color_list = [cmap(i) for i in np.linspace(0, 1, len(target_mzs))]
    else:
        color_list = colors

    for i, target_mz in enumerate(target_mzs):
        plt.plot(np.array(times), xics[target_mz],
                 label=f'$\it{{m/z}}$ {target_mz:.2f}', color=color_list[i % len(color_list)], linewidth=linewidth)

    plt.xlabel('RT (min)', fontname='Arial', fontsize=12, labelpad=4, color='0.2')
    # plt.ylabel('Intensity', fontname='Arial', fontsize=12, labelpad=2, color='0.2')

    if legend:
        plt.legend(prop={'family': 'Arial', 'size': 10})

    if rt_range:
        plt.xlim(rt_range)

    ax = plt.gca()

    # Remove top, right, and left spines
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)

    # Set the color of the frame (axes spines)
    for spine in ax.spines.values():
        spine.set_color('0.5')

    ax.yaxis.set_visible(False)

    if not x_axis:
        ax.xaxis.set_visible(False)
        for spine in ax.spines.values():
            spine.set_visible(True)

    # Set the color of tick labels to 0.2
    ax.tick_params(axis='both', colors='0.2')

    if y_max:
        plt.ylim(0, y_max)

    plt.tight_layout()

    if save:
        plt.savefig(name, transparent=True)
    plt.show()


from scipy.signal import savgol_filter
def smooth_xics(xics, window_size=3):
    """
    Function to smooth XICs using a moving average.
    :param xics: Dictionary of XICs.
    :param window: Window size for the moving average.
    :return: Dictionary of smoothed XICs.
    """
    smoothed_xics = {}
    for target_mz, intensities in xics.items():
        smoothed_xics[target_mz] = savgol_filter(intensities, window_size, 2)
    return smoothed_xics


if __name__ == '__main__':

    #############
    target_mzs = [103.06, 107.05, 120.08, 131.05, 149.06, 166.09]

    times, xics = get_xics('/data/nist_samples/data_1/Nist_pool_2_0eV.mzML',
                           target_mzs)

    # smoothing
    xics = smooth_xics(xics, window_size=3)

    plot_xics(times, xics, target_mzs, rt_range=(0.8, 1.1), y_max=None, x_axis=True, legend=False,
              save=False, name='xics1.svg')
    #
    # plot_xics(times, xics, target_mzs, rt_range=(0.8, 1.05), y_max=3e6, x_axis=False, legend=False,
    #           save=True, name='xics2.svg')

    plot_xics(times, xics, target_mzs, rt_range=(0.8, 1.05), y_max=3e6, x_axis=False, legend=True,
              save=False, name='xics3.svg')

    #############
    # target_mzs = [238.1071, 123.0443, 165.0544, 182.0811, 136.0757]
    #
    # times, xics = get_xics('/Users/shipei/Documents/projects/ms1_id/data/nist_samples/data_1/Nist_pool_1_10eV.mzML',
    #                        target_mzs)
    # plot_xics(times, xics, target_mzs, fig_size=(10, 6), linewidth=1.5,
    #           rt_range=(0.6, 0.8), y_max=5e7, x_axis=True, legend=True,
    #           save=False, name='xics_.svg')



import pickle
import matplotlib.pyplot as plt
import numpy as np
import matplotlib.colors as mcolors

# Load data
ms2_inchikeys = set(pickle.load(open('data/ms2_inchikey_gnps.pkl', 'rb')))
ms1_dda_inchikeys = set(pickle.load(open('data/ms1_inchikey_gnps_DDA.pkl', 'rb')))
ms1_0ev_inchikeys = set(pickle.load(open('data/ms1_inchikey_gnps_0ev.pkl', 'rb')))
ms1_10ev_inchikeys = set(pickle.load(open('data/ms1_inchikey_gnps_10ev.pkl', 'rb')))
ms1_20ev_inchikeys = set(pickle.load(open('data/ms1_inchikey_gnps_20ev.pkl', 'rb')))

# Calculate overlaps
ms2_overlap = [
    len(ms2_inchikeys.intersection(ms2_inchikeys)),
    len(ms2_inchikeys.intersection(ms1_dda_inchikeys)),
    len(ms2_inchikeys.intersection(ms1_0ev_inchikeys)),
    len(ms2_inchikeys.intersection(ms1_10ev_inchikeys)),
    len(ms2_inchikeys.intersection(ms1_20ev_inchikeys))
]

total_counts = [
    len(ms2_inchikeys),
    len(ms1_dda_inchikeys),
    len(ms1_0ev_inchikeys),
    len(ms1_10ev_inchikeys),
    len(ms1_20ev_inchikeys)
]

non_overlap = [total - overlap for total, overlap in zip(total_counts, ms2_overlap)]

# Set up the plot
plt.rcParams['font.family'] = 'Arial'
fig, ax = plt.subplots(figsize=(4, 3))

# Create the stacked bars
bar_width = 0.65
index = np.arange(5)

colors = ['#56648a', '#6280a5', '#8ca5c0', '#8d7e95', '#ca9a96', '#facaa9']
# Add transparency
overlap_color = mcolors.to_rgba(colors[2], alpha=0.7)
non_overlap_color = mcolors.to_rgba(colors[4], alpha=0.7)

p1 = ax.bar(index, ms2_overlap, bar_width, color=overlap_color, label='Overlap with MS/MS annotations')
p2 = ax.bar(index[1:], non_overlap[1:], bar_width, bottom=ms2_overlap[1:], color=non_overlap_color, label='Non-overlap')

# Customize the plot
ax.set_ylabel('Annotated metabolites', fontsize=12, color='0.2')
ax.set_xticks(index)
ax.set_xticklabels(('MS/MS\n(DDA)', 'MS1\n(DDA)', 'MS1\n(0 eV)',
                    'MS1\n(10 eV)', 'MS1\n(20 eV)'), fontsize=10, color='0.15')
# ax.legend(fontsize=10)

# Adjust frame
ax.spines['top'].set_visible(False)
ax.spines['right'].set_visible(False)
ax.spines['left'].set_color('0.4')
ax.spines['bottom'].set_color('0.4')

# Adjust tick length and pad
ax.tick_params(axis='x', which='major', labelsize=10, length=2.5, width=1, pad=3.5, color='0.4', labelcolor='0.2')
ax.tick_params(axis='y', which='major', labelsize=10, length=2.5, width=1, pad=1.5, color='0.4', labelcolor='0.2')


def add_labels(rects):
    for rect in rects:
        height = rect.get_height()
        ax.annotate(f'{height:,}',
                    xy=(rect.get_x() + rect.get_width() / 2, rect.get_y() + height / 2),
                    xytext=(0, 0),
                    textcoords="offset points",
                    ha='center', va='center', fontsize=9, color='0.1')


if __name__ == '__main__':
    add_labels(p1)
    add_labels(p2)

    # Adjust layout and display
    plt.tight_layout()
    # plt.savefig('data/ms_data_comparison_bar_plot.svg', transparent=True,
    #             bbox_inches='tight')
    plt.show()
